import logging
# from version import __version__


# initalize logger
logger = logging.getLogger('parqueryd')
ch = logging.StreamHandler()
formatter = logging.Formatter('%(levelname)s %(asctime)s %(name)s %(message)s', datefmt="%Y-%m-%d %H:%M:%S")
ch.setFormatter(formatter)
logger.addHandler(ch)

# import helpers



import base64
import json
import pickle
import time


def msg_factory(msg):
    if type(msg) is str:
        try:
            msg = json.loads(msg)
        except:
            msg is None
    if not msg:
        return Message()
    msg_mapping = {'calc': CalcMessage, 'rpc': RPCMessage, 'error': ErrorMessage,
                   'worker_register': WorkerRegisterMessage,
                   'busy': BusyMessage, 'done': DoneMessage,
                   'ticketdone': TicketDoneMessage,
                   'stop': StopMessage, None: Message}
    try:
        msg_class = msg_mapping.get(msg.get('msg_type'))
    except:
        print(msg)
        raise
    return msg_class(msg)


class MalformedMessage(Exception):
    pass


class Message(dict):
    msg_type = None

    def __init__(self, datadict=None):
        if datadict is None:
            datadict = {}
        self.update(datadict)
        self['payload'] = datadict.get('payload')
        self['version'] = datadict.get('version', 1)
        self['msg_type'] = self.msg_type
        self['created'] = time.time()

    def copy(self):
        newme = super(Message, self).copy()
        return msg_factory(newme)

    def isa(self, payload_or_instance):
        if self.msg_type == getattr(payload_or_instance, 'msg_type', '_'):
            return True
        if self.get('payload') == payload_or_instance:
            return True
        return False

    def add_as_binary(self, key, value):
        self[key] = base64.b64encode(pickle.dumps(value)).decode()

    def get_from_binary(self, key, default=None):
        buf = self.get(key)
        if not buf:
            return default
        return pickle.loads(base64.b64decode(buf))

    def to_json(self):
        # We could do some serializiation fixes in here for things like datetime or other binary non-json-serializabe members
        return json.dumps(self)

    def set_args_kwargs(self, args, kwargs):
        params = {'args': args, 'kwargs': kwargs}
        self.add_as_binary('params', params)

    def get_args_kwargs(self):
        params = self.get_from_binary('params', {})
        kwargs = params.get('kwargs', {})
        args = params.get('args', [])
        return args, kwargs


class WorkerRegisterMessage(Message):
    msg_type = 'worker_register'


class CalcMessage(Message):
    msg_type = 'calc'


class RPCMessage(Message):
    msg_type = 'rpc'


class ErrorMessage(Message):
    msg_type = 'error'


class BusyMessage(Message):
    msg_type = 'busy'


class DoneMessage(Message):
    msg_type = 'done'


class StopMessage(Message):
    msg_type = 'stop'


class TicketDoneMessage(Message):
    msg_type = 'ticketdone'



import logging
import os
import shutil
import socket
import threading
import time
from contextlib import contextmanager
from time import sleep
from uuid import uuid4
import glob

import boto3
import numpy as np
import pandas as pd
import pytest
import redis
from parquery.write import df_to_parquet

import parqueryd
import parqueryd.config
from parqueryd.worker import DownloaderNode

TEST_REDIS = 'redis://redis:6379/0'
FAKE_ACCESS_KEY = 'fake access key'
FAKE_SECRET_KEY = 'fake secret key'


class LocalS3Downloader(DownloaderNode):
    """Extend ``parqueryd.DownloaderNode`` to return an S3 Resource connected to local (test) S3."""

    def __init__(self, *args, **kwargs):
        super(LocalS3Downloader, self).__init__(*args, **kwargs)

        session = boto3.session.Session(
            aws_access_key_id=FAKE_ACCESS_KEY,
            aws_secret_access_key=FAKE_SECRET_KEY,
            region_name='eu-west-1')

        self._conn = session.resource('s3', endpoint_url='http://localstack:4572')

    def _get_s3_conn(self):
        return FAKE_ACCESS_KEY, FAKE_SECRET_KEY, self._conn

    def _get_transport_params(self):
        return {
            'resource_kwargs':
                {'endpoint_url': 'http://localstack:4572'}
        }


@pytest.fixture
def redis_server():
    """
    Return a Redis Client connected to local (test) redis.
    Remove all keys from REDIS before and after use.
    """
    redis_server = redis.from_url(TEST_REDIS)
    redis_server.flushdb()
    yield redis_server
    redis_server.flushdb()


@pytest.fixture
def clear_incoming():
    if os.path.isdir(parqueryd.config.INCOMING):
        shutil.rmtree(parqueryd.config.INCOMING)
    os.makedirs(parqueryd.config.INCOMING)


@pytest.fixture
def downloader():
    downloader = LocalS3Downloader(redis_url=TEST_REDIS, loglevel=logging.DEBUG)
    downloader_thread = threading.Thread(target=downloader.go)
    downloader_thread.daemon = True
    downloader_thread.start()
    # Sleep 5 seconds, just to make sure all connections are properly established
    sleep(5)

    yield downloader

    # shutdown the downloader
    downloader.running = False


@contextmanager
def clean_bucket(s3_conn, bucket_name):
    bucket = s3_conn.create_bucket(Bucket=bucket_name)
    for item in bucket.objects.all():
        item.delete()

    yield bucket

    for item in bucket.objects.all():
        item.delete()

    bucket.delete()


@pytest.mark.usefixtures('clear_incoming')
def test_downloader(redis_server, downloader, tmpdir):
    # Make a parquet from a pandas DataFrame
    data_df = pd.DataFrame(
        data=np.random.rand(100, 10),
        columns=['col_{}'.format(i + 1) for i in range(10)])
    local_parquet = str(tmpdir.join('test.parquet'))
    df_to_parquet(data_df, local_parquet)

    assert os.path.exists(local_parquet)

    s3_conn = downloader._get_s3_conn()[-1]

    with clean_bucket(s3_conn, 'parquet') as bucket:
        bucket.put_object(Key='test.parquet', Body=open(local_parquet, 'rb'))

        uploads = [key.key for key in bucket.objects.all()]
        assert uploads == ['test.parquet']

        # Construct the redis entry that the downloader is looking for
        progress_slot = '%s_%s' % (time.time() - 60, -1)
        node_filename_slot = '%s_%s' % (socket.gethostname(), 's3://parquet/test.parquet')
        ticket = str(uuid4())

        incoming_dir = parqueryd.config.INCOMING
        assert not glob.glob(os.path.join(incoming_dir, '*'))

        redis_server.hset(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot, progress_slot)

        # wait for the downloader to catch up
        sleep(10)

        # Check that incoming dir now has the test.parquet file.
        assert glob.glob(os.path.join(incoming_dir, '*')) == [os.path.join(incoming_dir, ticket + '_test.parquet')]

        # Check that the progress slot has been updated
        updated_slot = redis_server.hget(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot)
        assert updated_slot.split('_')[-1] == 'DONE'


if __name__ == '__main__':
    pytest.main([__file__, '-v', '-s'])



import logging
import os
import shutil
import socket
import threading
import time
from time import sleep
from uuid import uuid4

import numpy as np
import pandas as pd
import pytest
import redis
from parquery.write import df_to_parquet

import parqueryd.config
from parqueryd.worker import MoveparquetNode

TEST_REDIS = 'redis://redis:6379/0'


@pytest.fixture
def redis_server():
    """
    Return a Redis Client connected to local (test) redis.
    Remove all keys from REDIS before and after use.
    """
    redis_server = redis.from_url(TEST_REDIS)
    redis_server.flushdb()
    yield redis_server
    redis_server.flushdb()


@pytest.fixture
def clear_dirs():
    if os.path.isdir(parqueryd.config.INCOMING):
        shutil.rmtree(parqueryd.config.INCOMING)
    if os.path.isdir(parqueryd.config.DEFAULT_DATA_DIR):
        shutil.rmtree(parqueryd.config.DEFAULT_DATA_DIR)
    os.makedirs(parqueryd.config.DEFAULT_DATA_DIR)
    os.makedirs(parqueryd.config.INCOMING)

@pytest.fixture
def mover():
    mover = MoveparquetNode(redis_url=TEST_REDIS, loglevel=logging.DEBUG)
    mover_thread = threading.Thread(target=mover.go)
    mover_thread.daemon = True
    mover_thread.start()
    # Sleep 5 seconds, just to make sure all connections are properly established
    sleep(5)

    yield mover

    # shutdown the thread
    mover.running = False


@pytest.mark.usefixtures('clear_dirs')
@pytest.mark.usefixtures('mover')
def test_moveparquet(redis_server, tmpdir):
    # Make a parquet from a pandas DataFrame
    data_df = pd.DataFrame(
        data=np.random.rand(100, 10),
        columns=['col_{}'.format(i + 1) for i in range(10)])
    local_parquet = str(tmpdir.join('test_mover.parquet'))
    df_to_parquet(data_df, local_parquet)

    assert os.path.exists(local_parquet)

    # copy the parquet directory to bqueyd.INCOMING
    ticket = str(uuid4())
    ticket_dir = os.path.join(parqueryd.config.INCOMING, ticket + '_test_mover.parquet')
    shutil.copy(local_parquet, ticket_dir)

    # Construct the redis entry that before downloading
    progress_slot = '%s_%s' % (time.time() - 60, -1)
    node_filename_slot = '%s_%s' % (socket.gethostname(), 's3://parquet/test_mover.parquet')

    redis_server.hset(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot, progress_slot)

    # wait for some time
    sleep(5)

    # At this stage, we don't expect the parquet directory to be moved to parqueryd.DEFAULT_DATA_DIR, because the progress
    # slot has not been updated yet
    files_in_default_data_dir = os.listdir(parqueryd.config.DEFAULT_DATA_DIR)
    files_in_default_data_dir.sort()
    assert files_in_default_data_dir == ['incoming']
    # ticket_dir still exists
    assert os.path.exists(ticket_dir)

    # Now update progress slot
    new_progress_slot = '%s_%s' % (time.time(), 'DONE')
    redis_server.hset(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot, new_progress_slot)

    # Sleep again
    sleep(5)
    files_in_default_data_dir = os.listdir(parqueryd.config.DEFAULT_DATA_DIR)
    files_in_default_data_dir.sort()
    assert files_in_default_data_dir == ['incoming', 'test_mover.parquet', 'test_mover.parquet.metadata']
    # ticket_dir should have been deleted.
    assert not os.path.exists(ticket_dir)


if __name__ == '__main__':
    pytest.main([__file__, '-v', '-s'])



import errno
import os
import shutil

def ens_bytes(inp):
    return bytes(inp.encode("utf-8"))

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def rm_file_or_dir(path):
    if path is not None and os.path.exists(path):
        if os.path.isdir(path):
            if os.path.islink(path):
                os.unlink(path)
            else:
                shutil.rmtree(path)
        else:
            if os.path.islink(path):
                os.unlink(path)
            else:
                os.remove(path)



import binascii
import os
import random
import time

import netifaces
import zmq


def get_my_ip():
    eth_interfaces = sorted(
        [ifname for ifname in netifaces.interfaces() if (ifname.startswith('eth') or ifname.startswith('en'))])
    if len(eth_interfaces) < 1:
        ifname = 'lo'
    else:
        ifname = eth_interfaces[-1]
    for x in netifaces.ifaddresses(ifname)[netifaces.AF_INET]:
        # Return first addr found
        return x['addr']


def bind_to_random_port(socket, addr, min_port=49152, max_port=65536, max_tries=100):
    "We can't just use the zmq.Socket.bind_to_random_port, as we wan't to set the identity before binding"
    for i in range(max_tries):
        try:
            port = random.randrange(min_port, max_port)
            socket.identity = bytes(('%s:%s' % (addr, port)).encode("utf-8"))
            socket.bind(bytes(('tcp://*:%s' % port).encode("utf-8")))
            # socket.bind('%s:%s' % (addr, port))
        except zmq.ZMQError as exception:
            en = exception.errno
            if en == zmq.EADDRINUSE:
                continue
            else:
                raise
        else:
            return socket.identity
    raise zmq.ZMQBindError("Could not bind socket to random port.")


def tree_checksum(path):
    allfilenames = set()
    for root, dirs, filenames in os.walk(path):
        for filename in filenames:
            allfilenames.add(os.path.join(root, filename))
    buf = ''.join(sorted(allfilenames))
    return hex(binascii.crc32(buf) & 0xffffffff)


###################################################################################################
# Various Utility methods for user-friendly info display

def show_workers(info_data, only_busy=False):
    'For the given info_data dict, show a humand-friendly overview of the current workers'
    nodes = {}
    for w in info_data.get('workers', {}).values():
        nodes.setdefault(w['node'], []).append(w)
    for k, n in nodes.items():
        print(k)
        for nn in n:
            if only_busy and not nn.get('busy'):
                continue
            print('   ', time.ctime(nn['last_seen']), nn.get('busy'))



import binascii
import json
import logging
import os
import pickle
import random
import time
import base64

import redis
import zmq
from parquery.aggregate import aggregate_pa
from parquery.transport import deserialize_pa_table
from pyarrow import ArrowInvalid

import parqueryd.config
from parqueryd.messages import msg_factory, RPCMessage, ErrorMessage

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


class RPCError(Exception):
    """Base class for exceptions in this module."""
    pass


class RPC(object):

    def __init__(self, address=None, timeout=120, redis_url='redis://127.0.0.1:6379/0', loglevel=logging.INFO,
                 retries=3):
        self.logger = parqueryd.logger.getChild('rpc')
        self.logger.setLevel(loglevel)
        self.context = zmq.Context()
        self.redis_url = redis_url
        redis_server = redis.from_url(redis_url)
        self.retries = retries
        self.timeout = timeout
        self.identity = binascii.hexlify(os.urandom(8))

        if not address:
            # Bind to a random controller
            controllers = list(redis_server.smembers(parqueryd.config.REDIS_SET_KEY))
            if len(controllers) < 1:
                raise Exception('No Controllers found in Redis set: ' + parqueryd.config.REDIS_SET_KEY)
            random.shuffle(controllers)
        else:
            controllers = [address]
        self.controllers = controllers
        self.connect_socket()

    def connect_socket(self):
        reply = None
        for c in self.controllers:
            self.logger.debug('Establishing socket connection to %s', c)
            tmp_sock = self.context.socket(zmq.REQ)
            tmp_sock.setsockopt(zmq.RCVTIMEO, 2000)
            tmp_sock.setsockopt(zmq.LINGER, 0)
            tmp_sock.identity = self.identity
            tmp_sock.connect(c)
            # first ping the controller to see if it responds at all
            msg = RPCMessage({'payload': 'ping'})
            tmp_sock.send_json(msg)
            try:
                reply = msg_factory(tmp_sock.recv_json())
                self.address = c
                break
            except:
                logging.exception("Unable to connect to %s", c)
                continue
        if reply:
            # Now set the timeout to the actual requested
            self.logger.debug("Connection OK, setting network timeout to %s milliseconds", self.timeout * 1000)
            self.controller = tmp_sock
            self.controller.setsockopt(zmq.RCVTIMEO, self.timeout * 1000)
        else:
            raise Exception('No controller connection')

    def __getattr__(self, name):

        def _rpc(*args, **kwargs):
            self.logger.debug('Call %s on %s' % (name, self.address))
            start_time = time.time()
            params = {}
            if args:
                params['args'] = args
            if kwargs:
                params['kwargs'] = kwargs

            # We do not want string args to be converted into unicode by the JSON machinery
            msg = RPCMessage({'payload': name})
            msg.add_as_binary('params', params)
            rep = None
            for x in range(self.retries):
                try:
                    self.controller.send_json(msg)
                    rep = self.controller.recv()
                    break
                except Exception as e:
                    self.controller.close()
                    self.logger.critical(e)
                    if x == self.retries:
                        raise e
                    else:
                        self.logger.debug("Error, retrying %s" % (x + 1))
                        self.connect_socket()
                        pass
            if not rep:
                raise RPCError("No response from DQE, retries %s exceeded" % self.retries)
            try:
                if name == 'groupby':
                    _, groupby_col_list, agg_list, where_terms_list = args[0], args[1], args[2], args[3]
                    result = self.uncompress_groupby_to_pq(rep, groupby_col_list, agg_list, where_terms_list,
                                                           aggregate=kwargs.get('aggregate', False))
                else:
                    rep = msg_factory(json.loads(rep))
                    result = rep.get_from_binary('result')
            except (ValueError, TypeError):
                self.logger.exception('Could not use RPC method: {}/{}'.format(name, rep))
                result = rep
            if isinstance(rep, ErrorMessage):
                raise RPCError(rep.get('payload'))
            stop_time = time.time()
            self.last_call_duration = stop_time - start_time
            return result

        return _rpc

    def uncompress_groupby_to_pq(self, result, groupby_col_list, agg_list, where_terms_list, aggregate=False):
        # uncompress result returned by the groupby and convert it to a Pandas DataFrame
        try:
            pa_table = deserialize_pa_table(result)
        except ArrowInvalid:
            # if it's not a pyarrow table, an error must have happened and we should have a string message
            try:
                raise ValueError(pickle.loads(base64.b64decode(result)))
            except:
                raise

        result_df = aggregate_pa(
            pa_table,
            groupby_col_list,
            agg_list,
            data_filter=None,  # we can assume the filtering already happened
            aggregate=aggregate)

        del pa_table
        return result_df

    def get_download_data(self):
        redis_server = redis.from_url(self.redis_url)
        tickets = set(redis_server.keys(parqueryd.config.REDIS_TICKET_KEY_PREFIX + '*'))
        data = {}
        for ticket in tickets:
            tmp = redis_server.hgetall(ticket)
            data[ticket] = tmp
        return data

    def downloads(self):
        data = self.get_download_data()
        buf = []
        for k, v in data.items():
            done_count = 0
            for kk, vv in v.items():
                if vv.endswith('_DONE'):
                    done_count += 1
            buf.append((k, '%s/%s' % (done_count, len(v))))
        return buf

    def delete_download(self, ticket):
        redis_server = redis.from_url(self.redis_url)
        tmp = redis_server.hgetall(ticket)
        count = 0
        for k, v in tmp.items():
            count += redis_server.hdel(ticket, k)
        return count



#!/usr/bin/env python
import logging
import sys

import configobj

import parqueryd.config
from parqueryd.controller import ControllerNode
from parqueryd.rpc import RPC
from parqueryd.worker import WorkerNode, DownloaderNode, MoveparquetNode

config = configobj.ConfigObj('/etc/parqueryd.cfg')

redis_url = config.get('redis_url', 'redis://127.0.0.1:6379/0')
azure_conn_string = config.get('azure_conn_string', None)


def main(argv=sys.argv):
    if '-vvv' in argv:
        loglevel = logging.DEBUG
    elif '-vv' in argv:
        loglevel = logging.INFO
    elif '-v' in argv:
        loglevel = logging.WARNING
    else:
        loglevel = logging.ERROR

    data_dir = parqueryd.config.DEFAULT_DATA_DIR
    for arg in argv:
        if arg.startswith('--data_dir='):
            data_dir = arg[11:]

    if 'controller' in argv:
        ControllerNode(redis_url=redis_url, loglevel=loglevel, azure_conn_string=azure_conn_string).go()
    elif 'worker' in argv:
        WorkerNode(redis_url=redis_url, loglevel=loglevel, data_dir=data_dir).go()
    elif 'downloader' in argv:
        DownloaderNode(redis_url=redis_url, loglevel=loglevel, azure_conn_string=azure_conn_string).go()
    elif 'moveparquet' in argv:
        MoveparquetNode(redis_url=redis_url, loglevel=loglevel).go()
    else:
        if len(argv) > 1 and argv[1].startswith('tcp:'):
            rpc = RPC(address=argv[1], redis_url=redis_url, loglevel=loglevel)
        else:
            rpc = RPC(redis_url=redis_url, loglevel=loglevel)
        import IPython
        IPython.embed()


if __name__ == '__main__':
    main()



import logging
import os
import threading
from time import sleep

import pandas as pd
import pytest
import redis
from pandas.util.testing import assert_frame_equal
from parquery.write import df_to_parquet

import parqueryd.config
from parqueryd.controller import ControllerNode
from parqueryd.rpc import RPC
from parqueryd.util import get_my_ip
from parqueryd.worker import WorkerNode, DownloaderNode

TEST_REDIS = 'redis://redis:6379/0'
NR_SHARDS = 5


@pytest.fixture(scope='module')
def taxi_df():
    taxi_df = pd.read_csv(
        '/srv/datasets/yellow_tripdata_2016-01.csv',
        parse_dates=['tpep_pickup_datetime', 'tpep_dropoff_datetime']
    )
    yield taxi_df


class DummyDownloader(DownloaderNode):

    def download_file(self, ticket, filename):
        self.file_downloader_progress(ticket, filename, 'DONE')


@pytest.fixture(scope='module')
def rpc():
    redis_server = redis.from_url(TEST_REDIS)
    redis_server.flushdb()
    controller = ControllerNode(redis_url=TEST_REDIS, loglevel=logging.DEBUG)
    controller_thread = threading.Thread(target=controller.go)
    controller_thread.daemon = True
    controller_thread.start()
    # Sleep 5 seconds, just to make sure all connections are properly established
    sleep(5)

    worker = WorkerNode(redis_url=TEST_REDIS, loglevel=logging.DEBUG,
                        restart_check=False)
    worker_thread = threading.Thread(target=worker.go)
    worker_thread.daemon = True
    worker_thread.start()
    # Sleep 5 seconds, just to make sure all connections are properly established
    sleep(5)

    downloader = DummyDownloader(redis_url=TEST_REDIS, loglevel=logging.DEBUG)
    downloader_thread = threading.Thread(target=downloader.go)
    downloader_thread.daemon = True
    downloader_thread.start()
    sleep(5)

    rpc = parqueryd.rpc.RPC(timeout=100, redis_url=TEST_REDIS, loglevel=logging.DEBUG)
    yield rpc

    # shutdown the controller and worker
    controller.running = False
    worker.running = False
    downloader.running = False


@pytest.fixture(scope='module')
def shards(taxi_df):
    shard_filenames = []
    single_parquet = os.path.join(parqueryd.config.DEFAULT_DATA_DIR, 'yellow_tripdata_2016-01.parquet')
    df_to_parquet(taxi_df, single_parquet)
    shard_filenames.append(single_parquet)

    NR_SHARDS = 10
    step = len(taxi_df) // NR_SHARDS
    remainder = len(taxi_df) - step * NR_SHARDS
    count = 0
    for idx in range(0, len(taxi_df), step):
        if count == NR_SHARDS - 1 and remainder >= 0:
            step += remainder
        elif count == NR_SHARDS:
            break

        shard_file = os.path.join(parqueryd.config.DEFAULT_DATA_DIR, 'yellow_tripdata_2016-01-%s.parquet' % count)
        df_to_parquet(taxi_df[idx:idx + step], shard_file)
        shard_filenames.append(shard_file)
        count += 1

    yield shard_filenames


def test_rpc_info(rpc):
    result = rpc.info()

    address = result['address']
    my_ip = get_my_ip()
    assert my_ip == address.split(':')[1].replace('//', '')

    workers = result['workers']
    assert len(workers) == 2
    worker_node = list(v for v in workers.values() if v['workertype'] == 'calc')[0]

    node = result['node']
    assert node == worker_node['node']

    downloader_node = list(v for v in workers.values() if v['workertype'] == 'download')[0]
    assert node == downloader_node['node']

    assert result['others'] == {}


def test_download(rpc):
    ticket_nr = rpc.download(filenames=['test_download.parquet'], bucket='parquet', wait=False)
    redis_server = redis.from_url(TEST_REDIS)
    download_entries = redis_server.hgetall(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket_nr)
    assert len(download_entries) == 1
    for key, value in download_entries.items():
        # filename
        assert '_'.join(key.split('_')[1:]) == "s3://parquet/test_download.parquet"
        # progress slot
        assert value.split('_')[-1] == "-1"


def test_compare_with_pandas_total_amount_sum(taxi_df, rpc, shards):
    compare_with_pandas(taxi_df, rpc, shards, 'payment_type', 'total_amount', 'sum')


def test_compare_with_pandas_passenger_count_sum(taxi_df, rpc, shards):
    compare_with_pandas(taxi_df, rpc, shards, 'payment_type', 'passenger_count', 'sum')


@pytest.mark.skip('Skipping now as parquery does not implement means well yet')
def test_compare_with_pandas_total_amount_mean(taxi_df, rpc, shards):
    compare_with_pandas(taxi_df, rpc, shards, 'payment_type', 'total_amount', 'mean')


@pytest.mark.skip('Skipping now as parquery does not implement counts well yet')
def test_compare_with_pandas_payment_type_count(taxi_df, rpc, shards):
    compare_with_pandas(taxi_df, rpc, shards, 'payment_type', 'passenger_count', 'count')


def compare_with_pandas(taxi_df, rpc, shards, group_col, agg_col, method):
    full = os.path.basename(shards[0])
    full_result = rpc.groupby([full], [group_col], [[agg_col, method, agg_col]], [])
    full_result = full_result.sort_values(by=group_col)
    full_result = full_result.reset_index(drop=True)

    gp = taxi_df.groupby(group_col, sort=True, as_index=False)[agg_col]
    if method == 'sum':
        pandas_result = gp.sum()
    elif method == 'mean':
        pandas_result = gp.mean()
    elif method == 'count':
        pandas_result = gp.count()
    else:
        assert False, "Unknown method: {}".format(method)

    pandas_result = pandas_result.sort_values(by=group_col)
    pandas_result = pandas_result.reset_index(drop=True)

    assert_frame_equal(full_result, pandas_result, check_less_precise=True)


def test_compare_full_with_shard(rpc, shards):
    shard_filenames = [os.path.basename(x) for x in shards]

    full, parts = shard_filenames[:1], shard_filenames[1:]
    full_result = rpc.groupby(full, ['payment_type'], [['passenger_count', 'sum', 'passenger_count']], [])
    parts_result = rpc.groupby(parts, ['payment_type'], [['passenger_count', 'sum', 'passenger_count']], [])

    assert isinstance(full_result, pd.DataFrame)
    assert isinstance(parts_result, pd.DataFrame)

    # This returns a single DataFrame with the results from each part pasted in it separately
    # so we need to use a further groupby to produce the same result as with the full parquet
    parts_result = parts_result.groupby('payment_type', sort=True).sum()
    full_result = full_result.set_index('payment_type').sort_index()

    assert_frame_equal(full_result, parts_result)


if __name__ == '__main__':
    pytest.main([__file__, '-s'])



import binascii
import datetime
import errno
import gc
import glob
import importlib
import json
import logging
import os
import random
import shutil
import signal
import socket
import time
import traceback
from ssl import SSLError

import boto3
import psutil
import redis
import smart_open
import zmq
from azure.storage.blob import BlobClient
from parquery.aggregate import aggregate_pq
from parquery.transport import serialize_pa_table

import parqueryd.config
from parqueryd.messages import msg_factory, WorkerRegisterMessage, ErrorMessage, BusyMessage, StopMessage, \
    DoneMessage, TicketDoneMessage
from parqueryd.tool import rm_file_or_dir, ens_bytes

DATA_FILE_EXTENSION = '.parquet'
# timeout in ms : how long to wait for network poll, this also affects frequency of seeing new controllers and datafiles
POLLING_TIMEOUT = 5000
# how often in seconds to send a WorkerRegisterMessage
WRM_DELAY = 20
MAX_MEMORY_KB = 2 * (2 ** 20)  # Max memory of 2GB, in Kilobytes
DOWNLOAD_DELAY = 5  # how often in seconds to check for downloads


class WorkerBase(object):
    def __init__(self, data_dir=parqueryd.config.DEFAULT_DATA_DIR, redis_url='redis://127.0.0.1:6379/0',
                 loglevel=logging.DEBUG,
                 restart_check=True, azure_conn_string=None):
        if not os.path.exists(data_dir) or not os.path.isdir(data_dir):
            raise Exception("Datadir %s is not a valid directory" % data_dir)
        self.worker_id = binascii.hexlify(os.urandom(8)).decode()
        self.node_name = socket.gethostname()
        self.data_dir = data_dir
        self.data_files = set()
        self.restart_check = restart_check
        context = zmq.Context()
        self.socket = context.socket(zmq.ROUTER)
        self.socket.setsockopt(zmq.LINGER, 500)
        self.socket.identity = bytes((self.worker_id).encode("utf-8"))
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN | zmq.POLLOUT)
        self.redis_server = redis.from_url(redis_url)
        self.controllers = {}  # Keep a dict of timestamps when you last spoke to controllers
        self.check_controllers()
        self.last_wrm = 0
        self.start_time = time.time()
        self.logger = parqueryd.logger.getChild('worker ' + str(self.worker_id))
        self.logger.setLevel(loglevel)
        self.msg_count = 0
        signal.signal(signal.SIGTERM, self.term_signal())
        self.running = False
        self.azure_conn_string = azure_conn_string

    def term_signal(self):
        def _signal_handler(signum, frame):
            self.logger.info("Received TERM signal, stopping.")
            self.running = False

        return _signal_handler

    def send(self, addr, msg):
        try:
            if 'data' in msg:
                data = msg['data']
                del msg['data']
                self.socket.send_multipart([addr, msg.to_json(), data])
            else:
                self.socket.send_multipart([addr, msg.to_json()])
        except zmq.ZMQError as ze:
            self.logger.critical("Problem with %s: %s" % (str(addr), ze))

    def check_controllers(self):
        # Check the Redis set of controllers to see if any new ones have appeared,
        # Also register with them if so.
        listed_controllers = list(self.redis_server.smembers(parqueryd.config.REDIS_SET_KEY))
        current_controllers = []
        new_controllers = []
        for k in list(self.controllers.keys()):
            if k not in listed_controllers:
                del self.controllers[k]
                self.socket.disconnect(k)
            else:
                current_controllers.append(k)

        new_controllers = [c for c in listed_controllers if c not in current_controllers]
        for controller_address in new_controllers:
            self.socket.connect(controller_address)
            self.controllers[controller_address] = {'last_seen': 0, 'last_sent': 0,
                                                    'address': controller_address.decode()}

    def check_datafiles(self):
        has_new_files = False
        replacement_data_files = set()
        for data_file in [filename for filename in os.listdir(self.data_dir) if filename.endswith(DATA_FILE_EXTENSION)]:
            if data_file not in self.data_files:
                has_new_files = True
            replacement_data_files.add(data_file)
        self.data_files = replacement_data_files
        return has_new_files

    def prepare_wrm(self):
        wrm = WorkerRegisterMessage()
        wrm['worker_id'] = self.worker_id
        wrm['node'] = self.node_name
        wrm['data_files'] = list(self.data_files)
        wrm['data_dir'] = self.data_dir
        wrm['controllers'] = list(self.controllers.values())
        wrm['uptime'] = int(time.time() - self.start_time)
        wrm['msg_count'] = self.msg_count
        wrm['pid'] = os.getpid()
        wrm['workertype'] = self.workertype
        return wrm

    def heartbeat(self):
        time.sleep(0.001)  # to prevent too tight loop
        since_last_wrm = time.time() - self.last_wrm
        if since_last_wrm > WRM_DELAY:
            self.check_controllers()
            has_new_files = self.check_datafiles()
            self.last_wrm = time.time()
            wrm = self.prepare_wrm()
            for controller, data in self.controllers.items():
                if has_new_files or (time.time() - data['last_seen'] > WRM_DELAY):
                    self.send(controller, wrm)
                    data['last_sent'] = time.time()
                    self.logger.debug("heartbeat to %s", data['address'])

    def handle_in(self):
        try:
            tmp = self.socket.recv_multipart()
        except zmq.Again:
            return
        if len(tmp) != 2:
            self.logger.critical('Received a msg with len != 2, something seriously wrong. ')
            return
        sender, msg_buf = tmp
        self.logger.info("Received message from sender %s", sender)
        msg = msg_factory(msg_buf)

        data = self.controllers.get(sender)
        if not data:
            self.logger.critical('Received a msg from %s - this is an unknown sender' % sender)
            return
        data['last_seen'] = time.time()
        self.logger.debug('Received from %s', sender)

        # TODO Notify Controllers that we are busy, no more messages to be sent
        # The above busy notification is not perfect as other messages might be on their way already
        # but for long-running queries it will at least ensure other controllers
        # don't try and overuse this node by filling up a queue
        busy_msg = BusyMessage()
        self.send_to_all(busy_msg)

        try:
            tmp = self.handle(msg)
        except Exception:
            tmp = ErrorMessage(msg)
            tmp['payload'] = traceback.format_exc()
            self.logger.exception("Unable to handle message [%s]", msg)
        if tmp:
            self.send(sender, tmp)

        self.send_to_all(DoneMessage())  # Send a DoneMessage to all controllers, this flags you as 'Done'. Duh

    def go(self):
        self.logger.info('Starting')
        self.running = True
        while self.running:
            self.heartbeat()
            for sock, event in self.poller.poll(timeout=POLLING_TIMEOUT):
                if event & zmq.POLLIN:
                    self.handle_in()

        # Also send a message to all your controllers, that you are stopping
        self.send_to_all(StopMessage())
        for k in self.controllers:
            self.socket.disconnect(k)

        self.logger.info('Stopping')

    def send_to_all(self, msg):
        for controller in self.controllers:
            self.send(controller, msg)

    def handle(self, msg):
        if msg.isa('kill'):
            self.running = False
            return
        elif msg.isa('info'):
            msg = self.prepare_wrm()
        elif msg.isa('loglevel'):
            args, kwargs = msg.get_args_kwargs()
            if args:
                loglevel = {'info': logging.INFO, 'debug': logging.DEBUG}.get(args[0], logging.INFO)
                self.logger.setLevel(loglevel)
                self.logger.info("Set loglevel to %s" % loglevel)
            return
        elif msg.isa('readfile'):
            args, kwargs = msg.get_args_kwargs()
            msg['data'] = open(args[0]).read()
        elif msg.isa('sleep'):
            args, kwargs = msg.get_args_kwargs()
            time.sleep(float(args[0]))
            snore = 'z' * random.randint(1, 20)
            self.logger.debug(snore)
            msg.add_as_binary('result', snore)
        else:
            msg = self.handle_work(msg)
            self.msg_count += 1
            gc.collect()
            self._check_mem(msg)

        return msg

    def _check_mem(self, msg):
        # RSS is in bytes, convert to Kilobytes
        rss_kb = psutil.Process().memory_full_info().rss / (2 ** 10)
        self.logger.debug("RSS is: %s KB", rss_kb)
        if self.restart_check and rss_kb > MAX_MEMORY_KB:
            args = msg.get_args_kwargs()[0]
            self.logger.critical('args are: %s', args)
            self.logger.critical(
                'Memory usage (KB) %s > %s, restarting', rss_kb, MAX_MEMORY_KB)
            self.running = False

    def handle_work(self, msg):
        raise NotImplementedError


class WorkerNode(WorkerBase):
    workertype = 'calc'

    def execute_code(self, msg):
        args, kwargs = msg.get_args_kwargs()

        func_fully_qualified_name = kwargs['function'].split('.')
        module_name = '.'.join(func_fully_qualified_name[:-1])
        func_name = func_fully_qualified_name[-1]
        func_args = kwargs.get("args", [])
        func_kwargs = kwargs.get("kwargs", {})

        self.logger.debug('Importing module: %s' % module_name)
        mod = importlib.import_module(module_name)
        function = getattr(mod, func_name)
        self.logger.debug('Executing function: %s' % func_name)
        self.logger.debug('args: %s kwargs: %s' % (func_args, func_kwargs))
        result = function(*func_args, **func_kwargs)

        msg.add_as_binary('result', result)
        return msg

    def handle_work(self, msg):
        if msg.isa('execute_code'):
            return self.execute_code(msg)

        args, kwargs = msg.get_args_kwargs()
        self.logger.info('doing calc %s' % args)
        filename = args[0]
        groupby_col_list = args[1]
        aggregation_list = args[2]
        where_terms_list = args[3]
        aggregate = kwargs.get('aggregate', True)

        # create rootdir
        full_file_name = os.path.join(self.data_dir, filename)
        pa_table = aggregate_pq(
            full_file_name,
            groupby_col_list,
            aggregation_list,
            data_filter=where_terms_list,
            aggregate=aggregate,
            as_df=False
        )

        # create message
        if len(pa_table) == 0:
            msg['data'] = None
        else:
            msg['data'] = serialize_pa_table(pa_table)

        return msg


class DownloaderNode(WorkerBase):
    workertype = 'download'

    def __init__(self, *args, **kwargs):
        super(DownloaderNode, self).__init__(*args, **kwargs)
        self.last_download_check = 0

    def heartbeat(self):
        super(DownloaderNode, self).heartbeat()
        if time.time() - self.last_download_check > DOWNLOAD_DELAY:
            self.check_downloads()

    def check_downloads(self):
        # Note, the files being downloaded are stored per key on filename,
        # Yet files are grouped as being inside a ticket for downloading at the same time
        # this done so that a group of files can be synchronized in downloading
        # when called from rpc.download(filenames=[...]

        self.last_download_check = time.time()
        tickets = self.redis_server.keys(parqueryd.config.REDIS_TICKET_KEY_PREFIX + '*')

        for ticket_w_prefix in tickets:
            ticket_details = self.redis_server.hgetall(ticket_w_prefix)
            ticket = ticket_w_prefix[len(parqueryd.config.REDIS_TICKET_KEY_PREFIX):]

            ticket_details_items = [(k, v) for k, v in ticket_details.items()]
            random.shuffle(ticket_details_items)

            for node_filename_slot, progress_slot in ticket_details_items:
                tmp = node_filename_slot.split('_')
                if len(tmp) < 2:
                    self.logger.critical("Bogus node_filename_slot %s", node_filename_slot)
                    continue
                node = tmp[0]
                filename = '_'.join(tmp[1:])

                tmp = progress_slot.split('_')
                if len(tmp) != 2:
                    self.logger.critical("Bogus progress_slot %s", progress_slot)
                    continue
                timestamp, progress = float(tmp[0]), tmp[1]

                if node != self.node_name:
                    continue

                # If every progress slot for this ticket is DONE, we can consider the whole ticket done
                if progress == 'DONE':
                    continue

                try:
                    # acquire a lock for this node_filename
                    lock_key = parqueryd.config.REDIS_DOWNLOAD_LOCK_PREFIX + self.node_name + ticket + filename
                    lock = self.redis_server.lock(lock_key, timeout=parqueryd.config.REDIS_DOWNLOAD_LOCK_DURATION)
                    if lock.acquire(False):
                        self.download_file(ticket, filename)
                        break  # break out of the loop so we don't end up staying in a large loop for giant tickets
                except:
                    self.logger.exception('Problem downloading %s %s', ticket, filename)
                    # Clean up the whole ticket if an error occured
                    self.remove_ticket(ticket)
                    break
                finally:
                    try:
                        lock.release()
                    except redis.lock.LockError:
                        pass

    def file_downloader_progress(self, ticket, filename, progress):
        node_filename_slot = '%s_%s' % (self.node_name, filename)
        # Check to see if the progress slot exists at all, if it does not exist this ticket has been cancelled
        # by some kind of intervention, stop the download and clean up.
        tmp = self.redis_server.hget(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot)
        if not tmp:
            # Clean up the whole ticket contents from disk
            ticket_path = os.path.join(parqueryd.config.INCOMING, ticket)
            self.logger.debug('Now removing entire ticket %s', ticket_path)
            for filename in glob.glob(ticket_path + '*'):
                rm_file_or_dir(os.path.join(parqueryd.config.INCOMING, filename))
            raise Exception("Ticket %s progress slot %s not found, aborting download" % (ticket, node_filename_slot))
        # A progress slot contains a timestamp_filesize
        progress_slot = '%s_%s' % (time.time(), progress)
        self.redis_server.hset(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot, progress_slot)

    def _get_transport_params(self):
        return {}

    def download_file(self, ticket, fileurl):
        if self.azure_conn_string:
            self._download_file_azure(ticket, fileurl)
        else:
            self._download_file_aws(ticket, fileurl)

    def _download_file_aws(self, ticket, fileurl):
        tmp = fileurl.replace('s3://', '').split('/')
        bucket = tmp[0]
        filename = '/'.join(tmp[1:])

        incoming_file = self._get_temp_name(ticket, filename)

        if os.path.exists(incoming_file):
            self.logger.info("%s exists, skipping download" % incoming_file)
            self.file_downloader_progress(ticket, fileurl, 'DONE')
        else:
            self.logger.info("Downloading ticket [%s], bucket [%s], filename [%s]" % (ticket, bucket, filename))

            access_key, secret_key, s3_conn = self._get_s3_conn()
            object_summary = s3_conn.Object(bucket, filename)
            size = object_summary.content_length

            key = 's3://{}:{}@{}/{}'.format(access_key, secret_key, bucket, filename)

            with open(incoming_file, 'wb') as fd:
                # See: https://github.com/RaRe-Technologies/smart_open/commit/a751b7575bfc5cc277ae176cecc46dbb109e47a4
                # Sometime we get timeout errors on the SSL connections
                for x in range(3):
                    try:
                        transport_params = self._get_transport_params()
                        with smart_open.open(key, 'rb', transport_params=transport_params) as fin:
                            buf = True
                            progress = 0
                            while buf:
                                buf = fin.read(pow(2, 20) * 16)  # Use a bigger buffer
                                fd.write(buf)
                                progress += len(buf)
                                self.file_downloader_progress(ticket, fileurl, size)
                        break
                    except SSLError as e:
                        if x == 2:
                            raise e
                        else:
                            pass

        self.logger.debug('Download done %s: %s', ticket, fileurl)
        self.file_downloader_progress(ticket, fileurl, 'DONE')

    def _get_prod_name(self, file_name):
        return os.path.join(parqueryd.config.DEFAULT_DATA_DIR, file_name)

    def _get_temp_name(self, ticket, file_name):
        return os.path.join(parqueryd.config.INCOMING, ticket + '_' + file_name)

    def _get_s3_conn(self):
        """Create a boto3 """
        session = boto3.Session()
        credentials = session.get_credentials()
        if not credentials:
            raise ValueError('Missing S3 credentials')
        credentials = credentials.get_frozen_credentials()
        access_key = credentials.access_key
        secret_key = credentials.secret_key
        s3_conn = boto3.resource('s3')
        return access_key, secret_key, s3_conn

    def _download_file_azure(self, ticket, fileurl):
        tmp = fileurl.replace('azure://', '').split('/')
        container_name = tmp[0]
        blob_name = tmp[1]
        incoming_file = self._get_temp_name(ticket, blob_name)

        if os.path.exists(incoming_file):
            self.logger.info("%s exists, skipping download" % incoming_file)
            self.file_downloader_progress(ticket, fileurl, 'DONE')
        else:
            self.logger.info(
                "Downloading ticket [%s], container name [%s], blob name [%s]" % (ticket, container_name, blob_name))

            # Download blob
            with open(incoming_file, 'wb') as fh:
                blob_client = BlobClient.from_connection_string(
                    conn_str=self.azure_conn_string, container_name=container_name, blob_name=blob_name
                )
                download_stream = blob_client.download_blob()
                fh.write(download_stream.readall())

            self.logger.debug('Download done %s azure://%s/%s', ticket, container_name, blob_name)
            self.file_downloader_progress(ticket, fileurl, 'DONE')

    def remove_ticket(self, ticket):
        # Remove all Redis entries for this node and ticket
        # it can't be done per file as we don't have the bucket name from which a file was downloaded
        self.logger.debug('Removing ticket %s from redis', ticket)
        for node_filename_slot in self.redis_server.hgetall(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket):
            if node_filename_slot.startswith(self.node_name):
                self.logger.debug('Removing ticket_%s %s', ticket, node_filename_slot)
                self.redis_server.hdel(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot)
        tdm = TicketDoneMessage({'ticket': ticket})
        self.send_to_all(tdm)


class MoveparquetNode(DownloaderNode):
    workertype = 'moveparquet'

    def moveparquet(self, ticket):
        # A notification from the controller that all files are downloaded on all nodes,
        # the files in this ticket can be moved into place
        ticket_path = os.path.join(parqueryd.config.INCOMING, ticket + '_*')
        file_name_list = glob.glob(ticket_path)
        if file_name_list:
            for filename in file_name_list:
                filename_without_ticket = filename[filename.index('_') + 1:]
                prod_path = self._get_prod_name(filename_without_ticket)
                if os.path.exists(prod_path):
                    rm_file_or_dir(prod_path)
                incoming_path = self._get_temp_name(ticket, filename_without_ticket)

                # Add a metadata file to the downloaded item
                metadata_filepath = self._get_prod_name(filename_without_ticket + '.metadata')
                metadata = {'ticket': ticket,
                            'timestamp': time.time(),
                            'localtime': time.ctime(),
                            'utc': str(datetime.datetime.utcnow())
                            }
                open(metadata_filepath, 'w').write(json.dumps(metadata, indent=2))

                self.logger.debug("Moving %s %s" % (incoming_path, prod_path))
                shutil.move(incoming_path, prod_path)
        else:
            self.logger.debug('Doing a moveparquet for files %s which do not exist', ticket_path)

    def check_downloads(self):
        # Check all the entries for a specific ticket over all the nodes
        # only if ALL the nodes are _DONE downloading, move the parquet files in this ticket into place.

        self.last_download_check = time.time()
        tickets = self.redis_server.keys(parqueryd.config.REDIS_TICKET_KEY_PREFIX + '*')

        for ticket_w_prefix in tickets:
            ticket_details = self.redis_server.hgetall(ticket_w_prefix)
            ticket = ticket_w_prefix[len(parqueryd.config.REDIS_TICKET_KEY_PREFIX):]

            in_progress_count = 0
            ticket_details_items = ticket_details.items()

            ticket_on_this_node = False

            for node_filename_slot, progress_slot in ticket_details_items:
                tmp = node_filename_slot.split('_')
                if len(tmp) < 2:
                    self.logger.critical("Bogus node_filename_slot %s", node_filename_slot)
                    continue
                node = tmp[0]
                filename = '_'.join(tmp[1:])

                tmp = progress_slot.split('_')
                if len(tmp) != 2:
                    self.logger.critical("Bogus progress_slot %s", progress_slot)
                    continue
                timestamp, progress = float(tmp[0]), tmp[1]

                # If every progress slot for this ticket is DONE, we can consider the whole ticket done
                if progress != 'DONE':
                    in_progress_count += 1

                if node == self.node_name:
                    ticket_on_this_node = True

            if in_progress_count == 0 and ticket_on_this_node:
                try:
                    self.moveparquet(ticket)
                except:
                    self.logger.exception('Error occurred in moveparquet %s', ticket)
                finally:
                    self.remove_ticket(ticket)



from __future__ import absolute_import

########################################################################
#
# License: MIT
# Created: August 8, 2020
#       Author:  Carst Vaartjes - cvaartjes@visualfabriq.com
#
########################################################################
import codecs
import os

from setuptools import setup, Extension, find_packages
from os.path import abspath
from sys import version_info as v
from setuptools.command.build_ext import build_ext as _build_ext


# Check this Python version is supported
if any([v < (2, 6), (3,) < v < (3, 5)]):
    raise Exception("Unsupported Python version %d.%d. Requires Python >= 2.7 "
                    "or >= 3.5." % v[:2])


class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


HERE = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


def get_version():
    version = {}
    with open("parqueryd/version.py") as fp:
        exec (fp.read(), version)
    return version


# Sources & libraries
inc_dirs = [abspath('parqueryd')]
try:
    import numpy as np
    inc_dirs.append(np.get_include())
except ImportError as e:
    pass
lib_dirs = []
libs = []
def_macros = []
sources = []

cmdclass = {'build_ext': build_ext}

optional_libs = ['numexpr>=2.6.9']

install_requires = [
    'parquery>=0.1.11',
    'pyzmq>=19.0.2',
    'redis>=3.5.3',
    'boto3>=1.14.40',
    'netifaces>=0.10.9',
    'configobj>=5.0.6',
    'psutil>=5.7.2',
]
if v < (3,):
    install_requires.extend([
    'smart_open==1.10.1',
    'azure-storage-blob==12.3.2',
])
else:
    install_requires.extend([
    'smart_open>=1.11.1',
    'azure-storage-blob>=12.4.0',
])
setup_requires = []
tests_requires = [
    'pytest>=4.6.11',
    'pytest-cov>=2.10.0',
    'codacy-coverage>=1.3.11',
]
extras_requires = []
ext_modules = []
package_data = {}
classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
]

setup(
    name="parqueryd",
    version=get_version()['__version__'],
    description='A distribution framework for parquery',
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    classifiers=classifiers,
    author='Carst Vaartjes',
    author_email='cvaartjes@visualfabriq.com',
    maintainer='Carst Vaartjes',
    maintainer_email='cvaartjes@visualfabriq.com',
    url='https://github.com/visualfabriq/parqueryd',
    license='GPL2',
    platforms=['any'],
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_requires,
    extras_require=dict(
        optional=extras_requires,
        test=tests_requires
    ),
    packages=find_packages(),
    package_data=package_data,
    include_package_data=True,
    zip_safe=True,
    entry_points={
        'console_scripts': [
            'parqueryd = parqueryd.node:main'
        ]
    }
)




import os

# set standard data directory
DEFAULT_DATA_DIR = '/srv/parquet/'
if not os.path.exists(DEFAULT_DATA_DIR):
    os.makedirs(DEFAULT_DATA_DIR)
INCOMING = os.path.join(DEFAULT_DATA_DIR, 'incoming')
if not os.path.exists(INCOMING):
    os.makedirs(INCOMING)

# set the redis standards
REDIS_SET_KEY = 'parqueryd_controllers'
REDIS_TICKET_KEY_PREFIX = 'parqueryd_download_ticket_'
REDIS_DOWNLOAD_LOCK_PREFIX = 'parqueryd_download_lock_'
REDIS_DOWNLOAD_LOCK_DURATION = 60 * 30  # time in seconds to keep a lock



import os, traceback
import redis

REDIS_HOST = '127.0.0.1'
REDIS_PORT = 6379
REDIS_PREFIX = 'parqueryd_'

BROKER_PORT = 14334
VENTILATOR_PORT = 14335
SINK_PORT = 14336
RPC_PORT = 14337

DATA_FILE_EXTENSION = '.parquet'

try:
    if os.path.exists('config_local.py'):
        from config_local import *
except:
    traceback.print_exc()

REDIS = redis.Redis(host=REDIS_HOST, port=REDIS_PORT)

def get(key, default=None):
    tmp = REDIS.get(REDIS_PREFIX+key)
    return tmp or default


def set(key, value):
    REDIS.set(REDIS_PREFIX+key, value)


def append(key, value):
    'Append a value to a list on key'
    REDIS.rpush(REDIS_PREFIX+key, value)


def show_all():
    'Show all settings related to ' + REDIS_PREFIX
    pass


import binascii
import gc
import logging
import os
import random
import socket
import time
import traceback

import pyarrow as pa
import redis
import zmq
from parquery.transport import deserialize_pa_table, serialize_pa_table

import parqueryd
import parqueryd.config
from parqueryd.messages import msg_factory, Message, WorkerRegisterMessage, ErrorMessage, \
    BusyMessage, DoneMessage, StopMessage, TicketDoneMessage
from parqueryd.util import get_my_ip, bind_to_random_port
from parqueryd.tool import ens_bytes


POLLING_TIMEOUT = 500  # timeout in ms : how long to wait for network poll, this also affects frequency of seeing new nodes
DEAD_WORKER_TIMEOUT = 60  # time in seconds that we wait for a worker to respond before being removed
HEARTBEAT_INTERVAL = 2  # time in seconds between doing heartbeats
MIN_CALCWORKER_COUNT = 0.1  # percentage of workers that should ONLY do calcs and never do downloads to prevent download swamping
RUNFILES_LOCATION = '/srv/'  # Location to write a .pid and .address file


class ControllerNode(object):
    def __init__(self, redis_url='redis://127.0.0.1:6379/0', loglevel=logging.INFO, azure_conn_string=None):

        self.redis_url = redis_url
        self.redis_server = redis.from_url(redis_url)
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.setsockopt(zmq.LINGER, 500)
        self.socket.setsockopt(zmq.ROUTER_MANDATORY, 1)  # Paranoid for debugging purposes
        self.socket.setsockopt(zmq.SNDTIMEO, 1000)  # Short timeout
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN | zmq.POLLOUT)

        self.node_name = socket.gethostname()
        self.address = ens_bytes(bind_to_random_port(self.socket, 'tcp://' + get_my_ip(),
                                           min_port=14300, max_port=14399, max_tries=100).decode())
        with open(os.path.join(RUNFILES_LOCATION, 'parqueryd_controller.address'), 'w') as F:
            F.write(str(self.address))
        with open(os.path.join(RUNFILES_LOCATION, 'parqueryd_controller.pid'), 'w') as F:
            F.write(str(os.getpid()))

        self.logger = parqueryd.logger.getChild('controller').getChild(str(self.address))
        self.logger.setLevel(loglevel)

        self.msg_count_in = 0
        self.rpc_results = []  # buffer of results that are ready to be returned to callers
        self.rpc_segments = {}  # Certain RPC calls get split and divided over workers, this dict tracks the original RPCs
        self.worker_map = {}  # maintain a list of connected workers TODO get rid of unresponsive ones...
        self.files_map = {}  # shows on which workers a file is available on
        self.worker_out_messages = {None: []}  # A dict of buffers, used to round-robin based on message affinity
        self.worker_out_messages_sequence = [None]  # used to round-robin the outgoing messages
        self.is_running = True
        self.last_heartbeat = 0
        self.others = {}  # A dict of other Controllers running on other DQE nodes
        self.start_time = time.time()
        self.azure_conn_string = azure_conn_string

    def send(self, addr, msg_buf, is_rpc=False):
        try:
            if addr == self.address:
                self.handle_peer(addr, msg_factory(msg_buf))
                return
            if is_rpc:
                tmp = [addr, '', msg_buf]
            else:
                tmp = [addr, msg_buf]
            self.socket.send_multipart(tmp)
        except zmq.ZMQError as ze:
            self.logger.critical("Problem with %s: %s" % (addr, ze))

    def connect_to_others(self):
        # Make sure own address is still registered
        self.redis_server.sadd(parqueryd.config.REDIS_SET_KEY, self.address)
        # get list of other running controllers and connect to them
        all_servers = self.redis_server.smembers(parqueryd.config.REDIS_SET_KEY)
        all_servers.remove(self.address)
        # Connect to new other controllers we don't know about yet
        for x in all_servers:
            if x not in self.others:
                self.logger.info('Connecting to %s', x)
                self.socket.connect(x)
                self.others[x] = {'connect_time': time.time()}
            else:
                msg = Message({'payload': 'info'})
                msg.add_as_binary('result', self.get_info())
                try:
                    self.socket.send_multipart([x, msg.to_json()])
                except zmq.error.ZMQError as e:
                    self.logger.critical('Removing %s due to %s' % (x, e))
                    self.redis_server.srem(parqueryd.config.REDIS_SET_KEY, x)
                    del self.others[x]
        # Disconnect from controllers not in current set
        for x in list(self.others.keys()):  # iterate over a copy of keys so we can remove entries
            if x not in all_servers:
                self.logger.critical('Disconnecting from %s' % x)
                try:
                    del self.others[x]
                    self.socket.disconnect(x)
                except zmq.error.ZMQError as e:
                    self.logger.exception(e)

    def heartbeat(self):
        if time.time() - self.last_heartbeat > HEARTBEAT_INTERVAL:
            self.connect_to_others()
            self.last_heartbeat = time.time()

    def find_free_worker(self, needs_local=False, filename=None):
        # Pick a random worker_id to send a message, TODO add some kind of load-balancing
        free_workers = []
        free_local_workers = []

        for worker_id, worker in self.worker_map.items():
            # ignore downloader workers
            if worker.get('workertype') in ('download', 'moveparquet'):
                continue

            if worker.get('busy'):
                continue

            if filename and (worker_id not in self.files_map.get(filename, [])):
                continue

            free_workers.append(worker_id)
            if needs_local and worker.get('node') != self.node_name:
                continue
            if worker.get('node') == self.node_name:
                free_local_workers.append(worker_id)
        # if there are no free workers at all, just bail
        if not free_workers:
            return None

        if needs_local:
            if free_local_workers:
                return random.choice(free_local_workers)
            else:
                return None

        return random.choice(free_workers)

    def process_sink_results(self):
        while self.rpc_results:
            msg = self.rpc_results.pop()
            # If this was a message to be combined, there should be a parent_token
            if 'parent_token' in msg:
                parent_token = msg['parent_token']
                if parent_token not in self.rpc_segments:
                    logging.debug('Orphaned msg segment %s probabably an error was raised elsewhere' % parent_token)
                    continue
                original_rpc = self.rpc_segments.get(parent_token)

                if isinstance(msg, ErrorMessage):
                    self.logger.debug('Error messsage %s' % msg.get('payload'))
                    # Delete this entire message segments, if it still exists
                    if parent_token in self.rpc_segments:
                        del self.rpc_segments[parent_token]
                        # If any of the segment workers return an error,
                        # Send the exception on to the calling RPC
                        msg['token'] = parent_token
                        msg_id = binascii.unhexlify(parent_token)
                        self.send(msg_id, msg.to_json(), is_rpc=True)
                    # and continue the processing
                    continue

                args, kwargs = msg.get_args_kwargs()
                filename = args[0]

                if msg.get('data'):
                    result_table = deserialize_pa_table(msg['data'])
                else:
                    result_table = None

                original_rpc['results'][filename] = result_table

                if len(original_rpc['results']) == len(original_rpc['filenames']):
                    # Check to see that there are no filenames with no Result yet
                    # TODO as soon as any workers gives an error abort the whole enchilada

                    # We have received all the segment, send a reply to RPC caller
                    del msg
                    msg = original_rpc['msg']

                    # if finished, aggregate the result to a combined arrow table
                    pa_tables = [result_table for result_table in original_rpc['results'].values() if
                                 result_table is not None]
                    if pa_tables:
                        msg['data'] = serialize_pa_table(pa.concat_tables(pa_tables))

                    del self.rpc_segments[parent_token]
                else:
                    # This was a segment result move on
                    continue

            msg_id = binascii.unhexlify(msg.get('token'))
            if 'data' in msg:
                self.send(msg_id, msg['data'], is_rpc=True)
                gc.collect()
            else:
                self.send(msg_id, msg.to_json(), is_rpc=True)
            self.logger.debug('RPC Msg handled: %s' % msg.get('payload', '?'))

    def handle_out(self):
        # If there have been new affinity keys added, rotate them
        for k, v in self.worker_out_messages.items():
            if v and (k not in self.worker_out_messages_sequence):
                self.worker_out_messages_sequence.append(k)

        # remove unused affinities

        remove_list = [x for x in self.worker_out_messages_sequence if x not in self.worker_out_messages]
        for affinity in remove_list:
            self.worker_out_messages_sequence.remove(affinity)

        if not self.worker_out_messages_sequence:
            # nothing to do
            return

        nextq_key = self.worker_out_messages_sequence.pop(0)
        nextq = self.worker_out_messages.get(nextq_key)
        self.worker_out_messages_sequence.append(nextq_key)
        if not nextq:
            if nextq_key:  # Only delete those for which there is an affinity, the None q is default and should stay
                del self.worker_out_messages[nextq_key]
            return  # the next buffer is empty just return and try the next one in the next round

        msg = nextq[0]
        worker_id = msg.get('worker_id')

        # Take into account on which worker a filename is available
        filename = msg.get('filename')

        # find a worker that is free
        if worker_id == '__needs_local__':
            worker_id = self.find_free_worker(needs_local=True, filename=filename)
        elif worker_id is None:
            worker_id = self.find_free_worker(filename=filename)

        if not worker_id:
            # self.logger.debug('No free workers at this time')
            return

        msg = nextq.pop(0)

        # TODO Add a tracking of which requests have been sent out to the worker, and do retries with timeouts
        self.worker_map[worker_id]['last_sent'] = time.time()
        self.worker_map[worker_id]['busy'] = True
        self.send(worker_id, msg.to_json())

    def handle_in(self):
        self.msg_count_in += 1
        data = self.socket.recv_multipart()
        binary, sender = None, None  # initialise outside for edge cases
        if len(data) == 3:
            if data[1] == '':  # This is a RPC call from a zmq.REQ socket
                sender, _blank, msg_buf = data
                self.handle_rpc(sender, msg_factory(msg_buf))
                return
            sender, msg_buf, binary = data
        elif len(data) == 2:  # This is an internode call from another zmq.ROUTER, a Controller or Worker
            sender, msg_buf = data
        msg = msg_factory(msg_buf)
        if binary:
            msg['data'] = binary
        if sender in self.others:
            self.handle_peer(sender, msg)
        else:
            self.handle_worker(sender, msg)

    def handle_peer(self, sender, msg):
        if msg.isa('loglevel'):
            args, kwargs = msg.get_args_kwargs()
            loglevel = {'info': logging.INFO, 'debug': logging.DEBUG}.get(args[0], logging.INFO)
            self.logger.setLevel(loglevel)
            self.logger.info("Set loglevel to %s" % loglevel)
        elif msg.isa('info'):
            # Another node registered with you and is sending some info
            data = msg.get_from_binary('result', default={})
            addr = data.get('address')
            node = data.get('node')
            uptime = data.get('uptime')
            if addr and node:
                self.others[addr]['node'] = node
                self.others[addr]['uptime'] = uptime
            else:
                self.logger.critical("bogus Info message received from %s %s", sender, msg)
        else:
            self.logger.debug("Got a msg but don't know what to do with it %s" % msg)

    def handle_worker(self, worker_id, msg):

        # TODO Make a distinction on the kind of message received and act accordingly
        msg['worker_id'] = worker_id

        # TODO If worker not in worker_map due to a dead_worker cull (calculation too a long time...)
        # request a new WorkerRegisterMessage from that worker...
        if not msg.isa(WorkerRegisterMessage) and worker_id not in self.worker_map:
            self.send(worker_id, Message({'payload': 'info'}).to_json())

        self.worker_map.setdefault(worker_id, {})['last_seen'] = time.time()

        if msg.isa(WorkerRegisterMessage):
            self.logger.debug('Worker registered %s', worker_id)
            for filename in msg.get('data_files', []):
                self.files_map.setdefault(filename, set()).add(worker_id)
            self.worker_map[worker_id]['node'] = msg.get('node', '...')
            self.worker_map[worker_id]['uptime'] = msg.get('uptime', 0)
            self.worker_map[worker_id]['pid'] = msg.get('pid', '?')
            self.worker_map[worker_id]['workertype'] = msg.get('workertype', '?')
            return

        if msg.isa(BusyMessage):
            self.logger.debug('Worker %s sent BusyMessage', worker_id)
            self.worker_map[worker_id]['busy'] = True
            return

        if msg.isa(DoneMessage):
            self.logger.debug('Worker %s sent DoneMessage', worker_id)
            self.worker_map[worker_id]['busy'] = False
            return

        if msg.isa(StopMessage):
            self.remove_worker(worker_id)
            return

        if msg.isa(TicketDoneMessage):
            # Go through download tickets and see if there are any local RPCs waiting for it
            # Check the ticket in this number, if it is in the self.rpc_segments[ticket] of this controller
            # there is a RPC call waiting for it, so also answer that one

            ticket = msg.get('ticket')
            if ticket in self.rpc_segments:
                msg2 = self.rpc_segments[ticket]
                if 'token' in msg2:
                    # TODO Consider passing in some extra information to this ticket so the worker
                    # that did the downloading can pass in diagnostics etc.
                    self.rpc_results.append(msg2)
                del self.rpc_segments[ticket]
            return

        if 'token' in msg:
            # A message might have been passed on to a worker for processing and needs to be returned to
            #  the relevant caller so it goes in the rpc_results list
            self.rpc_results.append(msg)

    def handle_rpc(self, sender, msg):
        # RPC calls have a binary identiy set, hexlify it to make it readable and serializable
        msg_id = binascii.hexlify(sender)
        msg['token'] = msg_id
        # self.logger.debug('RPC received %s' % msg_id)

        result = "Sorry, I don't understand you"
        if msg.isa('ping'):
            result = 'pong'
        elif msg.isa('loglevel'):
            args, kwargs = msg.get_args_kwargs()
            if not args:
                result = "You need to specify a loglevel as first arg"
            else:
                m = msg.copy()
                del m['token']
                for x in self.others:
                    self.send(x, m.to_json())
                for x in self.worker_map:
                    self.send(x, m.to_json())
                self.handle_peer(None, m)
                result = "OK, loglevel set passed along"
        elif msg.isa('info'):
            result = self.get_info()
        elif msg.isa('kill'):
            result = self.kill()
        elif msg.isa('killworkers'):
            result = self.killworkers()
        elif msg.isa('killall'):
            result = self.killall()
        elif msg.isa('download'):
            result = self.setup_download(msg)
        elif msg.isa('readfile'):
            self.worker_out_messages[None].append(msg.copy())
            result = None
        elif msg.isa('execute_code'):
            args, kwargs = msg.get_args_kwargs()
            if 'function' not in kwargs:
                result = 'Error, function kwarg is missing!'
            else:
                self.worker_out_messages[None].append(msg.copy())
                if kwargs.get('wait', False):
                    result = None
                else:
                    result = 'OK, %s dispatched' % kwargs['function']
        elif msg['payload'] in ('sleep',):
            args, kwargs = msg.get_args_kwargs()
            if args:
                if type(args[0]) is int:
                    self.worker_out_messages[None].append(msg.copy())
                    result = None
                else:
                    for x in args[0]:
                        mm = msg.copy()
                        del mm['token']
                        affinity = kwargs.get('affinity')
                        mm.set_args_kwargs([x], kwargs)
                        self.worker_out_messages.setdefault(affinity, []).append(mm)
                    result = 'Multi-sleep returning immediately'
            else:
                result = "Sleep needs an int or list of ints as arg[0]"
        elif msg['payload'] in ('groupby',):
            result = self.handle_calc_message(
                msg)  # if result is not None something happened, return to caller immediately

        if result:
            msg.add_as_binary('result', result)
            self.rpc_results.append(msg)

    def setup_download(self, msg):
        args, kwargs = msg.get_args_kwargs()
        filenames = kwargs.get('filenames')
        bucket = kwargs.get('bucket')
        wait = kwargs.get('wait', False)
        if not (filenames and bucket):
            return "A download needs kwargs: (filenames=, bucket=)"

        # Turn filenames into URLs
        if self.azure_conn_string:
            filenames = ['azure://%s/%s' % (bucket, filename) for filename in filenames]
        else:
            filenames = ['s3://%s/%s' % (bucket, filename) for filename in filenames]

        ticket = binascii.hexlify(os.urandom(8))  # track all downloads using a ticket
        for filename in filenames:

            # get all node names from others + self
            # TODO if the download happens to start when you are disconnected from others this is an issue...
            nodes = [x.get('node') for x in self.others.values()]
            nodes.append(self.node_name)

            for node in nodes:
                # A progress slot contains a timestamp_filesize
                progress_slot = '%s_%s' % (
                    time.time() - 60,
                    -1)  # give the slot a timestamp of now -1 minute so we can see when it was created
                node_filename_slot = '%s_%s' % (node, filename)
                self.redis_server.hset(parqueryd.config.REDIS_TICKET_KEY_PREFIX + ticket, node_filename_slot,
                                       progress_slot)

        if wait:
            msg.add_as_binary('result', ticket)
            self.rpc_segments[ticket] = msg
            return None

        return ticket

    def handle_calc_message(self, msg):
        args, kwargs = msg.get_args_kwargs()
        affinity = kwargs.get('affinity')

        if len(args) != 4:
            return 'Error, No correct args given, expecting: ' + \
                   'path_list, groupby_col_list, measure_col_list, where_terms_list'

        filenames = args[0]
        if not filenames:
            return 'Error, no filenames given'

        # Make sure that all filenames are available before any messages are sent
        for filename in filenames:
            if filename and filename not in self.files_map:
                return 'Sorry, filename %s was not found' % filename

        parent_token = msg['token']
        rpc_segment = {'msg': msg_factory(msg.copy()),
                       'results': {},
                       'filenames': dict([(x, None) for x in filenames])}

        params = {}
        for filename in filenames:
            msg['filename'] = filename
            params['args'] = list(args)
            params['args'][0] = filename
            params['kwargs'] = kwargs
            msg.add_as_binary('params', params)

            # Make up a new token for the message sent to the workers, and collect the responses using that id
            msg['parent_token'] = parent_token
            new_token = binascii.hexlify(os.urandom(8))
            msg['token'] = new_token
            rpc_segment['filenames'][filename] = new_token
            self.worker_out_messages.setdefault(affinity, []).append(msg.copy())

        self.rpc_segments[parent_token] = rpc_segment

    def killall(self):
        self.killworkers()
        m = Message({'payload': 'kill'})
        for x in self.others:
            self.send(x, m.to_json(), is_rpc=True)
        self.kill()
        return 'dood'

    def kill(self):
        # unregister with the Redis set
        self.redis_server.srem(parqueryd.config.REDIS_SET_KEY, self.address)
        self.is_running = False
        return 'harakiri...'

    def killworkers(self):
        # Send a kill message to each of our workers
        for x in self.worker_map:
            self.send(x, Message({'payload': 'kill', 'token': 'kill'}).to_json())
        return 'hai!'

    def get_info(self):
        data = {'msg_count_in': self.msg_count_in, 'node': self.node_name,
                'workers': self.worker_map,
                'worker_out_messages': [(k, len(v)) for k, v in self.worker_out_messages.items()],
                'last_heartbeat': self.last_heartbeat, 'address': self.address.decode(),
                'others': {k.decode():v for k, v in self.others.items()},
                'rpc_results_len': len(self.rpc_results),
                'uptime': int(time.time() - self.start_time), 'start_time': self.start_time
                }
        return data

    def remove_worker(self, worker_id):
        self.logger.warning("Removing worker %s", worker_id)
        if worker_id in self.worker_map:
            del self.worker_map[worker_id]
        for worker_set in self.files_map.values():
            if worker_id in worker_set:
                worker_set.remove(worker_id)

    def free_dead_workers(self):
        now = time.time()
        for worker_id, worker in self.worker_map.items():
            if (now - worker.get('last_seen', now)) > DEAD_WORKER_TIMEOUT:
                self.remove_worker(worker_id)

    def go(self):
        self.logger.info('[#############################>. Starting .<#############################]')

        while self.is_running:
            try:
                time.sleep(0.001)
                self.heartbeat()
                self.free_dead_workers()
                for sock, event in self.poller.poll(timeout=POLLING_TIMEOUT):
                    if event & zmq.POLLIN:
                        self.handle_in()
                    if event & zmq.POLLOUT:
                        self.handle_out()
                self.process_sink_results()
            except KeyboardInterrupt:
                self.logger.debug('Keyboard Interrupt')
                self.kill()
            except:
                self.logger.error("Exception %s" % traceback.format_exc())

        self.logger.info('Stopping')
        for x in (os.path.join(RUNFILES_LOCATION, 'parqueryd_controller.pid'),
                  os.path.join(RUNFILES_LOCATION, 'parqueryd_controller.address')):
            if os.path.exists(x):
                os.remove(x)
